/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc> 

    This file is part of upnp-beacon.

    upnp-beacon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 as 
    published by the Free Software Foundation.

    upnp-beacon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with upnp-beacon. If not, see <http://www.gnu.org/licenses/>.

*/

#include "autoconfig.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <limits.h>
#include <libconfig.h>
#include <upnp.h>
#include <uuid/uuid.h>
#include <unistd.h>
#include <sys/stat.h>

#if defined(HAVE_GETOPT_H) && defined(HAVE_GETOPT_LONG)
    #include <getopt.h>
    #define OPTSTR "c:u:p:w:i:n:vh"
#else
    #error Need getopt_long!
#endif

#include "devdesc.h"
#include "ifutil.h"

#define UPNP_ALIVE_INTERVAL 1800

int g_shutdown_flag;

void signal_handler(int signum)
{
    if ((signum == SIGINT) || (signum == SIGTERM))
    {
        printf("\nShutting down...\n");
        g_shutdown_flag++;
    }
    return;
}

static void print_copyright()
{
    printf("\n%s version %s\n", PACKAGE_NAME, VERSION);
    printf("Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland\n\n");
}

static void print_usage(const char *myname)
{
    print_copyright();

    fprintf(stderr, "\n%s options\n\
\n\
Supported options:\n\
    --config or -c              use specified configuration file\n\
    --udnconfig or -u           use specified file for udn configuration\n\
    --port or -p                application webserver port\n\
    --interface or -i           network interface to bind to\n\
    --webroot or -w             use https instead of http in the URL\n\
    --changename or -n          write new name into config file and exit\n\
    --version                   print version information and exit\n\
    --help                      this help message\n\
\n", myname);
}

static void init_config_defaults(desc_config_t *desc_config)
{
    desc_config->url = NULL;
    desc_config->friendly_name = "UPnP Beacon";
    desc_config->manufacturer = "digitalSTROM.org";
    desc_config->manufacturer_url = "http://www.digitalstrom.org/";
    desc_config->modeldesc = "Allows discovery via UPnP";
    desc_config->modelname = "upnpbeacon";
    desc_config->udn = NULL;
    desc_config->ip = NULL;
    desc_config->protocol = "http";
    desc_config->port = NULL;
}

static int static_upnp_callback(Upnp_EventType eventtype, void *event, void *cookie)
{
    return UPNP_E_SUCCESS;
}

int main(int argc, char **argv)
{
    config_t config;
    config_t udn_config;
    config_setting_t *cfg_root;
    config_setting_t *setting;
    char *err = NULL;
    const char *interface = NULL;
    const char *cfg_file = "/etc/upnpbeacon.cfg";
    const char *udn_file = "/etc/upnpbeacon.udn";
    char *device_desc = NULL;
    UpnpDevice_Handle device_handle;
    desc_config_t desc_config;
    const char *webroot = PACKAGE_DATADIR "/www";
    int test_port;

    int port = -1;
    g_shutdown_flag = 0;
    uuid_t uuid;
    char uuid_str[37];

    struct sigaction action;

    int opt_index = 0;
    int o;
    int changename = 0;
    int ret;

    static struct option long_options[] = 
    {
        {"config",          1, 0, 'c'},
        {"port",            1, 0, 'p'},
        {"webroot",         1, 0, 'w'},
        {"interface",       1, 0, 'i'},
        {"changename",      1, 0, 'n'},
        {"version",         0, 0, 'v'},
        {"help",            0, 0, 'h'},
        {0, 0, 0, 0}
    };

    init_config_defaults(&desc_config);

    while (1)
    {
        o = getopt_long(argc, argv, OPTSTR, long_options, &opt_index);
        if (o == -1) break;

        switch (o)
        {
            case 'p':
                errno = 0;
                port = strtol(optarg, &err, 10);
                if ((port == 0) && (*err))
                {
                    fprintf(stderr, "Invalid application port argument: %s\n", 
                            optarg);
                    exit(EXIT_FAILURE);
                }

                if (port > USHRT_MAX)
                {
                    fprintf(stderr, "Invalid application port value %d. "
                            "Maximum allowed port value is %d\n", port, 
                            USHRT_MAX);
                    exit(EXIT_FAILURE);
                }
                break;
            case 'c':
                cfg_file = optarg;
                break;
            case 'u':
                udn_file = optarg;
                break;
            case 'w':
                webroot = optarg;
                break;
            case 'i':
                interface = optarg;
                break;
            case 'n':
                changename = 1;
                desc_config.friendly_name = optarg;
                break;
            case 'v':
                print_copyright();
                exit(EXIT_SUCCESS);
                break;
            case '?':
            case 'h':
            default:
                print_usage(argv[0]);
                exit(EXIT_FAILURE);
        }
    }

    print_copyright();

    struct stat statbuf;
    if (stat(cfg_file, &statbuf) != 0)
    {
        fprintf(stderr, "Could not find configuration file %s\n", cfg_file);
        exit(EXIT_FAILURE);
    }

    if (stat(udn_file, &statbuf) != 0)
    {
        FILE *f = fopen(udn_file, "w");
        if (f == NULL)
        {
            fprintf(stderr, "Could not create udn configuration file %s: %s\n", 
                    udn_file, strerror(errno));
            exit(EXIT_FAILURE);
        }
        fclose(f);
    }
    
    config_init(&config);
    if (!config_read_file(&config, cfg_file)) 
    {
        fprintf(stderr, "%s:%d - %s\n", cfg_file, 
                config_error_line(&config), config_error_text(&config));
        config_destroy(&config);
        exit(EXIT_FAILURE);
    }

    if (changename)
    {
        cfg_root = config_root_setting(&config);
        setting = config_setting_add(cfg_root, "upnp-name", CONFIG_TYPE_STRING);
        if (setting == NULL)
        {
            setting = config_setting_get_member(cfg_root, "upnp-name");
        }
        config_setting_set_string(setting, desc_config.friendly_name);
        ret = config_write_file(&config, cfg_file);
        config_destroy(&config);
        if (!ret)
        {
            fprintf(stderr, "Error while writing configuration file %s\n",
                    cfg_file);
            exit(EXIT_FAILURE);
        }

        exit(EXIT_SUCCESS);
    }
    else
    {
        config_lookup_string(&config, "upnp-name", &desc_config.friendly_name);
    }

    config_init(&udn_config);
    if (!config_read_file(&udn_config, udn_file)) 
    {
        fprintf(stderr, "%s:%d - %s\n", udn_file, 
                config_error_line(&udn_config), config_error_text(&udn_config));
        config_destroy(&udn_config);
        config_destroy(&config);
        exit(EXIT_FAILURE);
    }

    if (!config_lookup_string(&config, "presentation-url", &desc_config.url))
    {
        fprintf(stderr, "Could not find required parameter "
                        "\"presentation-url\" in configuration!\n");
        config_destroy(&config);
        config_destroy(&udn_config);
        exit(EXIT_FAILURE);
    }

    if (!config_lookup_string(&udn_config, "udn", &desc_config.udn) || 
        (strlen(desc_config.udn) == 0))
    {
        uuid_generate(uuid);
        uuid_unparse_lower(uuid, uuid_str);
        desc_config.udn = uuid_str;
        cfg_root = config_root_setting(&udn_config);
        setting = config_setting_add(cfg_root, "udn", CONFIG_TYPE_STRING);
      
        if (setting == NULL)
        {
            setting = config_setting_get_member(cfg_root, "udn");
        }
        config_setting_set_string(setting, uuid_str);
        if (!config_write_file(&udn_config, udn_file))
        {
            fprintf(stderr, "Error while writing configuration file %s\n",
                    udn_file);
            config_destroy(&config);
            config_destroy(&udn_config);
            exit(EXIT_FAILURE);
        }
    }

    config_lookup_string(&config, "manufacturer", &desc_config.manufacturer);
    config_lookup_string(&config, "manufacturer-url",
                                                &desc_config.manufacturer_url);
    config_lookup_string(&config, "model-description", &desc_config.modeldesc);
    config_lookup_string(&config, "model-name", &desc_config.modelname);
    if (!config_lookup_string(&config, "presentation-port", &desc_config.port))
    {
        fprintf(stderr, "Could not find required parameter "
                        "\"presentation-port\" in configuration!\n");
        config_destroy(&config);
        config_destroy(&udn_config);
        exit(EXIT_FAILURE);
    }

    if (interface != NULL)
    {
        desc_config.ip = interface_to_ip(interface);
    }

    errno = 0;
    test_port = strtol(desc_config.port, &err, 10);
    if ((test_port == 0) && (*err))
    {
        fprintf(stderr, "Invalid presentation URL port argument: %s\n", 
                  desc_config.port);
        config_destroy(&config);
        config_destroy(&udn_config);
        exit(EXIT_FAILURE);
    }
    
    if ((test_port < 0) || (test_port > USHRT_MAX))
    {
        fprintf(stderr, "Invalid port specified in configuration!\n");
        config_destroy(&config);
        config_destroy(&udn_config);
        exit(EXIT_FAILURE);
    }

    memset(&action, 0, sizeof(action));
    action.sa_handler = signal_handler;
    action.sa_flags = 0;
    sigfillset(&action.sa_mask);

    if (sigaction(SIGINT, &action, NULL) < 0)
    {
        fprintf(stderr, "Could not register SIGINT handler!\n");
        exit(EXIT_FAILURE);
    }

    if (sigaction(SIGTERM, &action, NULL) < 0)
    {
        fprintf(stderr, "Could not register SIGTERM handler!\n");
        exit(EXIT_FAILURE);
    }

    if (port < 0)
    {
        port = 0;
    }

    ret = UpnpInit(desc_config.ip, port);
    if (ret != UPNP_E_SUCCESS)
    {
        fprintf(stderr, "UpnpInit failed: %d\n", ret);
        free(desc_config.ip);
        config_destroy(&config);
        config_destroy(&udn_config);
        exit(EXIT_FAILURE);
    }

    if ((desc_config.ip == NULL) && (UpnpGetServerIpAddress() != NULL))
    {
        desc_config.ip = strdup(UpnpGetServerIpAddress());
    }

    device_desc = device_desc_from_config(&desc_config);
    if (device_desc == NULL)
    {
        free(desc_config.ip);
        config_destroy(&config);
        config_destroy(&udn_config);
        UpnpFinish();
        fprintf(stderr, "Could not generate device description document!\n");
        exit(EXIT_FAILURE);
    }

    ret = UpnpSetWebServerRootDir(webroot);
    if (ret != UPNP_E_SUCCESS)
    {
        fprintf(stderr, "Failed to set webroot directory!");
        free(desc_config.ip);
        free(device_desc);
        config_destroy(&config);
        config_destroy(&udn_config);
        UpnpFinish();
        exit(EXIT_FAILURE);
    }

    ret = UpnpRegisterRootDevice2(UPNPREG_BUF_DESC, device_desc,
                                  strlen(device_desc ) + 1, 1,
                                  static_upnp_callback,
                                  &device_handle,
                                  &device_handle);
    if (ret != UPNP_E_SUCCESS)
    {
        fprintf(stderr, "Failed to register UPnP root device!\n");
        free(desc_config.ip);
        free(device_desc);
        config_destroy(&config);
        config_destroy(&udn_config);
        UpnpFinish();
        exit(EXIT_FAILURE);
    }

    ret = UpnpUnRegisterRootDevice(device_handle);
    if (ret != UPNP_E_SUCCESS)
    {
        fprintf(stderr, "Failed to unregister UPnP root device!\n");
        free(desc_config.ip);
        free(device_desc);
        config_destroy(&config);
        config_destroy(&udn_config);
        UpnpFinish();
        exit(EXIT_FAILURE);
    }

    ret = UpnpRegisterRootDevice2(UPNPREG_BUF_DESC, device_desc,
                                  strlen(device_desc ) + 1, 1,
                                  static_upnp_callback,
                                  &device_handle,
                                  &device_handle);
    if (ret != UPNP_E_SUCCESS)
    {
        fprintf(stderr, "Failed to register UPnP root device!\n");
        free(desc_config.ip);
        free(device_desc);
        config_destroy(&config);
        config_destroy(&udn_config);
        UpnpFinish();
        exit(EXIT_FAILURE);
    }

    ret = UpnpSendAdvertisement(device_handle, UPNP_ALIVE_INTERVAL);
    if (ret != UPNP_E_SUCCESS)
    {
        fprintf(stderr, "Failed to send UPnP alive advertisements!\n");
        UpnpUnRegisterRootDevice(device_handle);
        free(desc_config.ip);
        free(device_desc);
        config_destroy(&config);
        config_destroy(&udn_config);
        UpnpFinish();
        exit(EXIT_FAILURE);
    }

    do
    {
        pause();
    } while (!g_shutdown_flag);

    UpnpUnRegisterRootDevice(device_handle);
    UpnpFinish();
    free(desc_config.ip);
    free(device_desc);
    config_destroy(&config);
    config_destroy(&udn_config);
    return EXIT_SUCCESS;
}
